#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['kalong', 'kalong.utils', 'kalong.utils.doc_lookup']

package_data = \
{'': ['*'],
 'kalong': ['assets/*',
            'front/*',
            'front/Answer/*',
            'front/Code/*',
            'front/Prompt/*',
            'front/font/*']}

install_requires = \
['aiohttp', 'jedi']

extras_require = \
{'dev': ['black', 'flake8', 'isort'],
 'extras': ['cutter', 'uncompyle6'],
 'test': ['flake8', 'pytest-cov', 'pytest-flake8', 'pytest-isort', 'pytest'],
 'tools': ['sphinx']}

setup(name='kalong',
      version='0.1.1',
      description='A new take on debugging',
      author='Florian Mounier',
      author_email='paradoxxx.zero@gmail.com',
      url='https://github.com/paradoxxxzero/kalong',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
     )
